/* Radeon - TVOut class for the Radeon cards
 *
 * Copyright (c) 2004-2005 pinc Software. All Rights Reserved.
 */


#include "TVOut.h"
#include "Utility.h"
#include "multimon.h"

#include <WindowScreen.h>
#include <Application.h>
#include <Screen.h>
#include <Bitmap.h>
#include <String.h>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>


class RadeonTVOut : public TVOut {
	public:
		RadeonTVOut(BMessage &settings);
		virtual ~RadeonTVOut();

		virtual status_t InitCheck();

		virtual status_t SetEnabled(bool enabled);
		virtual bool IsEnabled() const;

		virtual status_t SetMode(int32 mode);
		virtual status_t QueryModes(BMessage &modes);

		virtual void GetName(char *buffer, size_t bufferSize);
		virtual void GetInfo(char *buffer, size_t bufferSize);

	private:
		bool _GetDisplayMode(display_mode& displayMode);
		void _UpdateTVOut();

		bool 			fIsEnabled;
		uint32			fTVMode;
		display_mode	fOldMode;
		bool			fOldSwapDisplays;
		int32			fWorkspace;
};


static float
get_refresh_rate(display_mode& mode)
{
	// we have to be catious as refresh rate cannot be controlled directly,
	// so it suffers under rounding errors and hardware restrictions
	return rint(10 * float(mode.timing.pixel_clock * 1000) / 
		float(mode.timing.h_total * mode.timing.v_total)) / 10.0;
}


static BString
tv_standard_to_string(uint32 mode)
{
	switch (mode) {
		case 0:		return "disabled";
		case 1:		return "NTSC";
		case 2:		return "NTSC Japan";
		case 3:		return "PAL BDGHI";
		case 4:		return "PAL M";
		case 5:		return "PAL N";
		case 6:		return "SECAM";
		case 101:	return "NTSC 443";
		case 102:	return "PAL 60";
		case 103:	return "PAL NC";
		default:
		{
			BString name;
			name << "??? (" << mode << ")";

			return name;
		}
	}
}


//	#pragma mark -


RadeonTVOut::RadeonTVOut(BMessage &settings)
	: TVOut(settings),
	fIsEnabled(false)
{
	fTVMode = 3;
}


RadeonTVOut::~RadeonTVOut()
{
}


status_t 
RadeonTVOut::InitCheck()
{
}


void 
RadeonTVOut::_UpdateTVOut()
{
	BScreen screen;
	SetSwapDisplays(&screen, true);
	SetTVStandard(&screen, fTVMode);

	display_mode mode;
	if (_GetDisplayMode(mode))
		screen.SetMode(fWorkspace, &mode);
}


status_t 
RadeonTVOut::SetEnabled(bool enabled)
{
	if (fIsEnabled == enabled)
		return B_OK;

	status_t status = B_OK;

	if (fIsEnabled) {
		// restore original display mode
		BScreen screen;
		SetSwapDisplays(&screen, fOldSwapDisplays);
		SetTVStandard(&screen, 0);

		screen.SetMode(fWorkspace, &fOldMode);
	} else {
		// save current display mode
		BScreen screen;
		GetSwapDisplays(&screen, &fOldSwapDisplays);

		screen.GetMode(&fOldMode);
		fWorkspace = current_workspace();

		_UpdateTVOut();
	}

	if (status == B_OK)
		fIsEnabled = enabled;

	return status;
}


bool 
RadeonTVOut::IsEnabled() const
{
	return fIsEnabled;
}


status_t 
RadeonTVOut::SetMode(int32 mode)
{
	if (fTVMode == mode)
		return B_OK;

	fTVMode = mode;
	if (fIsEnabled)
		_UpdateTVOut();

	return B_OK;
}


status_t 
RadeonTVOut::QueryModes(BMessage &modes)
{
	// first, we iterate over the NTSC modes, then comes the rest
	int32 ntscModes = 0, index = 0;
	uint32 standard;
	BScreen screen;

	while (GetNthSupportedTVStandard(&screen, index++, &standard) == B_OK) {
		BString name = tv_standard_to_string(standard);
		if (name.ByteAt(0) != 'N')
			continue;

		BString label = "800x600 ";
		label << name;

		modes.AddString("name", label);
		modes.AddInt32("mode", standard);
		ntscModes++;
	}

	// now let's add all none-NTSC modes

	index = 0;
	while (GetNthSupportedTVStandard(&screen, index++, &standard) == B_OK) {
		if (standard == 0)
			continue;

		BString name = tv_standard_to_string(standard);
		if (name.ByteAt(0) == 'N')
			continue;

		BString label = "800x600 ";
		label << name;

		modes.AddString("name", label);
		modes.AddInt32("mode", standard);
	}

	modes.AddInt32("ntsc_modes", ntscModes);
	modes.AddInt32("current", fTVMode);

	return B_OK;
}


void 
RadeonTVOut::GetName(char *buffer, size_t bufferSize)
{
	// R5 has no strlcpy()
	strncpy(buffer, "Radeon", bufferSize - 1);
	buffer[bufferSize - 1] = '\0';
}


void 
RadeonTVOut::GetInfo(char *buffer, size_t bufferSize)
{
	accelerant_device_info info;
	BScreen screen;
	if (screen.GetDeviceInfo(&info) != B_OK)
		strcpy(info.name, "written by Thomas Kurschel");

	// R5 has no strlcpy()
	strncpy(buffer, info.name, bufferSize - 1);
	buffer[bufferSize - 1] = '\0';
}


bool
RadeonTVOut::_GetDisplayMode(display_mode& displayMode)
{
	float oldRefresh = get_refresh_rate(fOldMode);
	uint16 virtualWidth, virtualHeight;
	int32 bestIndex = -1;
	float bestDiff = 999;
	display_mode* modes;
	uint32 modeCount;

	BScreen screen;
	if (screen.GetModeList(&modes, &modeCount) != B_OK)
		return false;

	virtualWidth = 800;
	virtualHeight = 600;

	// try to find mode in list provided by driver
	for (uint32 i = 0; i < modeCount; i++) {
		if (modes[i].virtual_width != virtualWidth
			|| modes[i].virtual_height != virtualHeight
			|| (color_space)modes[i].space != fOldMode.space)
			continue;

		float refresh = get_refresh_rate(modes[i]);
		if (refresh == oldRefresh) {
			// we have luck - we can use this mode directly
			displayMode = modes[i];
			displayMode.h_display_start = 0;
			displayMode.v_display_start = 0;
			free(modes);
			return true;
		}

		float diff = fabs(refresh - oldRefresh);
		if (diff < bestDiff) {
			bestDiff = diff;
			bestIndex = i;
		}
	}

	// we didn't find the exact mode, but something very similar?
	if (bestIndex == -1) {
		free(modes);
		return false;
	}

	// now, we are better of using GMT formula, but
	// as we don't have it, we just tune the pixel
	// clock of the best mode.

	displayMode = modes[bestIndex];
	displayMode.h_display_start = 0;
	displayMode.v_display_start = 0;

	// after some fiddling, it looks like this is the formula
	// used by the original panel (notice that / 10 happens before
	// multiplying with refresh rate - this leads to different
	// rounding)
	displayMode.timing.pixel_clock = ((uint32)displayMode.timing.h_total
		* displayMode.timing.v_total / 10 * int32(oldRefresh * 10)) / 1000;

	free(modes);
	return true;
}

//	#pragma mark -


TVOut *
GetRadeonTVOut(BMessage &settings)
{
	// check for Radeon card and Thomas's driver API!
	// (as of now, we don't care if it's a Radeon, since it's the only card
	// that implements this interface, and will likely to stay alone)

	BScreen screen;
	uint32 tvStandard;
	if (GetNthSupportedTVStandard(&screen, 0, &tvStandard) != B_OK && tvStandard > 0)
		return NULL;

	return new RadeonTVOut(settings);
}

